@@ -1,47 +1,18 @@
 import ee
 import numpy as np
+import warnings
+from .common import _get_indices
+from .common import _get_platform
+from .common import _get_scale_method
+from .common import _get_expression_map
 
 def _extend_eeImageCollection():
     """Decorator. Extends the ee.ImageCollection class."""
     return lambda f: (setattr(ee.imagecollection.ImageCollection,f.__name__,f) or f)
 
-def _get_platform(collection):
-    '''Gets the platform (satellite) of an image collection and wheter if it is a Surface Reflectance product.
-    
-    Parameters
-    ----------    
-    collection : ee.Image.Collection
-        Image to get platform.
-        
-    Returns
-    -------
-    dict :
-        Platform and product of the image.
-    '''
-    platforms = [
-        'COPERNICUS/S2',
-        'LANDSAT/LC08',
-        'LANDSAT/LE07',
-        'LANDSAT/LT05',
-        'LANDSAT/LT04',
-    ]
-    
-    imgID = collection.first().get('system:id').getInfo()
-    plt = None
-    
-    for platform in platforms:
-        if platform in imgID:
-            plt = platform
-        if '_SR' in imgID:
-            platformDict = {'platform': plt, 'sr': True}
-        else:
-            platformDict = {'platform': plt, 'sr': False}
-            
-    return platformDict
-
 @_extend_eeImageCollection()
-def closest(self, date):
-    '''Gets the closest image (or set of images if the collection intersects a region that requires multiple scenes) closest to the specified date.
+def closest(self, date, tolerance = 1, unit = 'month'):
+    '''Gets the closest image (or set of images if the collection intersects a region that requires multiple scenes) to the specified date.
     
     Parameters
     ----------    
@@ -49,6 +20,11 @@ def closest(self, date):
         Image Collection from which to get the closest image to the specified date.
     date : ee.Date | string
         Date of interest. The method will look for images closest to this date.
+    tolerance : float, default = 1
+        Filter the collection to [date - tolerance, date + tolerance) before searching the closest image. This speeds up the searching process for collections
+        with a high temporal resolution.
+    unit : string, default = 'month'
+        Units for tolerance. Available units: 'year', 'month', 'week', 'day', 'hour', 'minute' or 'second'.
         
     Returns
     -------    
@@ -58,6 +34,10 @@ def closest(self, date):
     if not isinstance(date, ee.ee_date.Date):
         date = ee.Date(date)
     
+    startDate = date.advance(-tolerance, unit)
+    endDate = date.advance(tolerance, unit)
+    self = self.filterDate(startDate, endDate)
+    
     def setProperties(img):        
         img = img.set('dateDist',ee.Number(img.get('system:time_start')).subtract(date.millis()).abs())        
         img = img.set('day',ee.Date(img.get('system:time_start')).get('day')) 
@@ -81,23 +61,27 @@ def index(self,index = 'NDVI',G = 2.5,C1 = 6.0,C2 = 7.5,L = 1.0):
     Parameters
     ----------     
     self : ee.ImageCollection
-        Image collection to compute indices on. Must be scaled to [0,1].
+        Image collection to compute indices on. Must be scaled to [0,1]. Check the supported platforms in User Guide > Spectral Indices > Supported Platforms.
     index : string | list[string], default = 'NDVI'
         Index or list of indices to compute.\n
         Available options:
             - 'vegetation' : Compute all vegetation indices.
             - 'burn' : Compute all burn indices.
             - 'water' : Compute all water indices.
+            - 'snow' : Compute all snow indices.
             - 'all' : Compute all indices listed below.
         Vegetation indices:     
             - 'BNDVI' : Blue Normalized Difference Vegetation Index.
             - 'CIG' : Chlorophyll Index - Green.
             - 'CVI' : Chlorophyll Vegetation Index.
             - 'EVI' : Enhanced Vegetation Index.
+            - 'GBNDVI' : Green-Blue Normalized Difference Vegetation Index.
             - 'GNDVI' : Green Normalized Difference Vegetation Index.
+            - 'GRNDVI' : Green-Red Normalized Difference Vegetation Index.
+            - 'MNDVI' : Modified Normalized Difference Vegetation Index.
             - 'NDVI' : Normalized Difference Vegetation Index.
             - 'NGRDI' : Normalized Green Red Difference Index.
-            - 'SR' : Simple Ratio.
+            - 'RVI' : Ratio Vegetation Index.
             - 'SAVI' : Soil-Adjusted Vegetation Index.
         Burn and fire indices:       
             - 'BAI' : Burned Area Index.
@@ -105,7 +89,9 @@ def index(self,index = 'NDVI',G = 2.5,C1 = 6.0,C2 = 7.5,L = 1.0):
             - 'NBR' : Normalized Burn Ratio.
         Water indices:     
             - 'MNDWI' : Modified Normalized Difference Water Index.
-            - 'NDWI' : Normalized Difference Water Index.        
+            - 'NDWI' : Normalized Difference Water Index.
+        Snow indices:     
+            - 'NDSI' : Normalized Difference Snow Index.
     G : float, default = 2.5
         Gain factor. Used just for index = 'EVI'.
     C1 : float, default = 6.0
@@ -122,165 +108,42 @@ def index(self,index = 'NDVI',G = 2.5,C1 = 6.0,C2 = 7.5,L = 1.0):
     '''  
     platformDict = _get_platform(self)
     
-    def lookupDic(img):
-        
-        def lookupS2(img):
-            return {
-                'G': float(G),
-                'C1': float(C1),
-                'C2': float(C2),
-                'L': float(L),
-                'A': img.select('B1'),
-                'B': img.select('B2'),
-                'G': img.select('B3'),
-                'R': img.select('B4'),
-                'RE1': img.select('B5'),
-                'RE2': img.select('B6'),
-                'RE3': img.select('B7'),
-                'N' : img.select('B8'),
-                'RE4': img.select('B8A'),
-                'WV' : img.select('B9'),            
-                'S1': img.select('B11'),
-                'S2': img.select('B12')
-            }
-        
-        def lookupL8(img):
-            return {
-                'G': float(G),
-                'C1': float(C1),
-                'C2': float(C2),
-                'L': float(L),
-                'A': img.select('B1'),
-                'B': img.select('B2'),
-                'G': img.select('B3'),
-                'R': img.select('B4'),
-                'N': img.select('B5'),
-                'S1': img.select('B6'),
-                'S2': img.select('B7'),                
-                'T1' : img.select('B10'),
-                'T2': img.select('B11')
-            }
-        
-        def lookupL457(img):
-            return {
-                'G': float(G),
-                'C1': float(C1),
-                'C2': float(C2),
-                'L': float(L),
-                'B': img.select('B1'),
-                'G': img.select('B2'),
-                'R': img.select('B3'),
-                'N': img.select('B4'),
-                'S1': img.select('B5'),
-                'T1': img.select('B6'),
-                'S2': img.select('B7')                
-            }
-        
-        lookupPlatform = {
-            'COPERNICUS/S2': lookupS2,
-            'LANDSAT/LC08': lookupL8,
-            'LANDSAT/LE07': lookupL457,
-            'LANDSAT/LT05': lookupL457,
-            'LANDSAT/LT04': lookupL457
-        }
-        
-        return lookupPlatform[platformDict['platform']](img)
-    
-    # VEGETATION INDICES
-    
-    def BNDVI(img):
-        return img.addBands(img.expression('(N - B) / (N + B)',lookupDic(img)).rename('BNDVI'))
-    
-    def CIG(img):
-        return img.addBands(img.expression('(N / G) - 1',lookupDic(img)).rename('CIG'))
-    
-    def CVI(img):
-        return img.addBands(img.expression('(N * R) / (G ** 2)',lookupDic(img)).rename('CVI'))
-    
-    def EVI(img):
-        return img.addBands(img.expression('G * (N - R) / (N + C1 * R - C2 * B + L)',lookupDic(img)).rename('EVI'))
-       
-    def GNDVI(img):
-        return img.addBands(img.expression('(N - G) / (N + G)',lookupDic(img)).rename('GNDVI'))
-    
-    def NDVI(img):
-        return img.addBands(img.expression('(N - R) / (N + R)',lookupDic(img)).rename('NDVI'))
-    
-    def NGRDI(img):
-        return img.addBands(img.expression('(G - R) / (G + R)',lookupDic(img)).rename('NGRDI'))
-    
-    def SAVI(img):
-        return img.addBands(img.expression('(1 + L) * (N - R) / (N + R + L)',lookupDic(img)).rename('SAVI'))
-    
-    def SR(img):
-        return img.addBands(img.expression('N / R',lookupDic(img)).rename('SR'))
-    
-    lookupVegetation = {
-        'BNDVI': BNDVI,
-        'CIG': CIG,
-        'CVI': CVI,
-        'EVI': EVI,
-        'GNDVI': GNDVI,
-        'NDVI': NDVI,
-        'NGRDI': NGRDI,        
-        'SAVI': SAVI,
-        'SR': SR,
+    additionalParameters = {
+        'g': float(G),
+        'C1': float(C1),
+        'C2': float(C2),
+        'L': float(L),
     }
     
-    # BURN INDICES
-    
-    def BAI(img):
-        return img.addBands(img.expression('1.0 / ((0.1 - R) ** 2.0 + (0.06 - N) ** 2.0)',lookupDic(img)).rename('BAI'))
-    
-    def BAIS2(img):
-        first = img.expression('(1.0 - ((RE2 * RE3 * RE4) / R) ** 0.5)',lookupDic(img))
-        second = img.expression('(((S2 - RE4)/(S2 + RE4) ** 0.5) + 1.0)',lookupDic(img))
-        return img.addBands(first.multiply(second).rename('BAIS2'))
-    
-    def NBR(img):
-        return img.addBands(img.expression('(N - S2) / (N + S2)',lookupDic(img)).rename('NBR'))
-    
-    lookupBurn = {
-        'BAI': BAI,
-        'BAIS2': BAIS2,
-        'NBR': NBR
-    }
-    
-    # WATER INDICES
-        
-    def MNDWI(img):
-        return img.addBands(img.expression('(G - S1) / (G + S1)',lookupDic(img)).rename('MNDWI'))
-    
-    def NDWI(img):
-        return img.addBands(img.expression('(G - N) / (G + N)',lookupDic(img)).rename('NDWI'))
-    
-    lookupWater = {
-        'MNDWI': MNDWI,
-        'NDWI': NDWI,
-    }
-    
-    # ALL INDICES
-    
-    lookup = {**lookupVegetation, **lookupBurn, **lookupWater}
+    spectralIndices = _get_indices()
+    indicesNames = list(spectralIndices.keys())
     
     if not isinstance(index, list):
         if index == 'all':
-            index = list(lookup.keys())
-        elif index == 'vegetation':
-            index = list(lookupVegetation.keys())
-        elif index == 'burn':
-            index = list(lookupBurn.keys())
-        elif index == 'water':
-            index = list(lookupWater.keys())
+            index = list(spectralIndices.keys())
+        elif index in ['vegetation','burn','water','snow']:
+            temporalListOfIndices = []
+            for idx in indicesNames:
+                if spectralIndices[idx]['type'] == index:
+                    temporalListOfIndices.append(idx)
+            index = temporalListOfIndices
         else:
             index = [index]        
-    
-    if platformDict['platform'] != 'COPERNICUS/S2' and 'BAIS2' in index:
-        index.remove('BAIS2')
-    
+            
     for idx in index:
-        self = self.map(lookup[idx])
-    
+        if idx not in list(spectralIndices.keys()):
+            warnings.warn("Index " + idx + " is not a built-in index and it won't be computed!",Warning)
+        else:
+            def temporalIndex(img):
+                lookupDic = _get_expression_map(img, platformDict)
+                lookupDic = {**lookupDic, **additionalParameters}
+                if all(band in list(lookupDic.keys()) for band in spectralIndices[idx]['requires']):
+                    return img.addBands(img.expression(spectralIndices[idx]['formula'],lookupDic).rename(idx))                
+                else:
+                    warnings.warn("This platform doesn't have the required bands for " + idx + " computation!",Warning)
+                    return img
+            self = self.map(temporalIndex)
+            
     return self
 
 @_extend_eeImageCollection()
@@ -290,13 +153,7 @@ def maskClouds(self, method = 'cloud_prob', prob = 60, maskCirrus = True, maskSh
     Parameters
     ----------    
     self : ee.ImageCollection [this]
-        Image collection to mask.\n
-        Accepted platforms:
-            - Sentinel-2
-            - Landsat 8
-            - Landsat 7
-            - Landsat 5
-            - Landsat 4
+        Image collection to mask. Check the supported platforms in User Guide > Masking Clouds and Shadows > Supported Platforms.
     method : string, default = 'cloud_prob'
         Method used to mask clouds.\n
         Available options:
@@ -329,6 +186,11 @@ def maskClouds(self, method = 'cloud_prob', prob = 60, maskCirrus = True, maskSh
     ee.ImageCollection
         Cloud-shadow masked image collection.
     '''
+    def S3(args):
+        qa = args.select('quality_flags')
+        notCloud = qa.bitwiseAnd(1 << 27).eq(0);
+        return args.updateMask(notCloud)
+    
     def S2(args):    
 
         def cloud_prob(img):
@@ -409,16 +271,85 @@ def maskClouds(self, method = 'cloud_prob', prob = 60, maskCirrus = True, maskSh
         mask2 = args.mask().reduce(ee.Reducer.min());
         return args.updateMask(cloud.Not()).updateMask(mask2);
     
+    def MOD09GA(args):
+        qa = args.select('state_1km')
+        notCloud = qa.bitwiseAnd(1 << 0).eq(0)
+        if maskShadows:
+            notCloud = notCloud.And(qa.bitwiseAnd(1 << 2).eq(0))
+        if maskCirrus:
+            notCloud = notCloud.And(qa.bitwiseAnd(1 << 8).eq(0))        
+        return args.updateMask(notCloud)
+    
+    def MCD15A3H(args):
+        qa = args.select('FparExtra_QC')
+        notCloud = qa.bitwiseAnd(1 << 5).eq(0)
+        if maskShadows:
+            notCloud = notCloud.And(qa.bitwiseAnd(1 << 6).eq(0))
+        if maskCirrus:
+            notCloud = notCloud.And(qa.bitwiseAnd(1 << 4).eq(0))        
+        return args.updateMask(notCloud)
+    
+    def MOD09Q1(args):
+        qa = args.select('State')
+        notCloud = qa.bitwiseAnd(1 << 0).eq(0)
+        if maskShadows:
+            notCloud = notCloud.And(qa.bitwiseAnd(1 << 2).eq(0))
+        if maskCirrus:
+            notCloud = notCloud.And(qa.bitwiseAnd(1 << 8).eq(0))        
+        return args.updateMask(notCloud)
+    
+    def MOD09A1(args):
+        qa = args.select('StateQA')
+        notCloud = qa.bitwiseAnd(1 << 0).eq(0)
+        if maskShadows:
+            notCloud = notCloud.And(qa.bitwiseAnd(1 << 2).eq(0))
+        if maskCirrus:
+            notCloud = notCloud.And(qa.bitwiseAnd(1 << 8).eq(0))        
+        return args.updateMask(notCloud)
+    
+    def MOD17A2H(args):
+        qa = args.select('Psn_QC')
+        notCloud = qa.bitwiseAnd(1 << 3).eq(0)
+        return args.updateMask(notCloud)
+    
+    def MOD16A2(args):
+        qa = args.select('ET_QC')
+        notCloud = qa.bitwiseAnd(1 << 3).eq(0)
+        return args.updateMask(notCloud)
+    
+    def MOD13Q1A1(args):
+        qa = args.select('SummaryQA')
+        notCloud = qa.bitwiseAnd(1 << 0).eq(0)
+        return args.updateMask(notCloud)
+    
+    def MOD13A2(args):
+        qa = args.select('SummaryQA')
+        notCloud = qa.eq(0)
+        return args.updateMask(notCloud)
+    
     lookup = {
+        'COPERNICUS/S3': S3,
         'COPERNICUS/S2': S2,
         'LANDSAT/LC08': L8,
         'LANDSAT/LE07': L457,
         'LANDSAT/LT05': L457,
-        'LANDSAT/LT04': L457
+        'LANDSAT/LT04': L457,
+        'MODIS/006/MOD09GA': MOD09GA,
+        'MODIS/006/MCD15A3H': MCD15A3H,
+        'MODIS/006/MOD09Q1': MOD09Q1,
+        'MODIS/006/MOD09A1': MOD09A1,
+        'MODIS/006/MOD17A2H': MOD17A2H,
+        'MODIS/006/MOD16A2': MOD16A2,
+        'MODIS/006/MOD13Q1': MOD13Q1A1,
+        'MODIS/006/MOD13A1': MOD13Q1A1,
+        'MODIS/006/MOD13A2': MOD13A2
     }
     
     platformDict = _get_platform(self)
     
+    if platformDict['platform'] not in list(lookup.keys()):
+        raise Exception("Sorry, satellite platform not supported for cloud masking!")
+    
     if platformDict['platform'] == 'COPERNICUS/S2':
         maskedImageCollection = lookup[platformDict['platform']](self)
     else:
@@ -428,62 +359,24 @@ def maskClouds(self, method = 'cloud_prob', prob = 60, maskCirrus = True, maskSh
 
 @_extend_eeImageCollection()
 def scale(self):    
-    '''Scales reflectance bands on an image collection. This also scales temperature in Landsat images and derived bands in Sentinel-2 images
-    (except for MSK_CLDPRB and MSK_SNWPRB bands that are excluded from the image collection).
+    '''Scales bands on an image collection.
     
     Parameters
     ----------
     self : ee.ImageCollection (this)
-        Image collection to scale.\n
-        Accepted platforms:
-            - Sentinel-2
-            - Landsat 8
-            - Landsat 7
-            - Landsat 5
-            - Landsat 4
+        Image collection to scale. Check the supported platforms in User Guide > Image Scaling > Supported Platforms.
         
     Returns
     -------
     ee.ImageCollection
         Scaled image collection.
-    '''    
-    def S2(img):
-        scaled = img.select(['B.*']).divide(1e4)      
-        scaled = scaled.addBands(img.select(['Q.*']))
-        if platformDict['sr']:            
-            scaled = scaled.addBands(img.select(['AOT','WVP']).divide(1e3))
-            scaled = scaled.addBands(img.select(['T.*']))            
-            scaled = scaled.addBands(img.select('SCL'))
-        return ee.Image(scaled.copyProperties(img,img.propertyNames()))
+    '''
+    platformDict = _get_platform(self)
+    lookup = _get_scale_method(platformDict)
     
-    def L8(img):               
-        if platformDict['sr']:
-            scaled = img.select(['B[1-9]']).divide(1e4)
-            scaled = scaled.addBands(img.select(['B10','B11']).divide(10)) 
-            scaled = scaled.addBands(img.select(['sr_aerosol','pixel_qa','radsat_qa']))
-            return ee.Image(scaled.copyProperties(img,img.propertyNames()))
-        else:
-            pass
-        
-    def L457(img):               
-        if platformDict['sr']:
-            scaled = img.select(['B[1-5]','B7']).divide(1e4)
-            scaled = scaled.addBands(img.select(['B6']).divide(10)) 
-            scaled = scaled.addBands(img.select(['sr_atmos_opacity']).divide(1e3)) 
-            scaled = scaled.addBands(img.select(['sr_cloud_qa','pixel_qa','radsat_qa']))
-            return ee.Image(scaled.copyProperties(img,img.propertyNames()))
-        else:
-            pass
+    if platformDict['platform'] not in list(lookup.keys()):
+        raise Exception("Sorry, satellite platform not supported for scaling!")
     
-    lookup = {
-        'COPERNICUS/S2': S2,
-        'LANDSAT/LC08': L8,
-        'LANDSAT/LE07': L457,
-        'LANDSAT/LT05': L457,
-        'LANDSAT/LT04': L457
-    }
-    
-    platformDict = _get_platform(self)    
     scaledImageCollection = self.map(lookup[platformDict['platform']])
     
     return scaledImageCollection
\ No newline at end of file
